# Import submodule files so
# classes and functions are usable at
# 'from jutl.timers import _' level.
from .stopwatch import *
from .timer import *

# Only show functions specified in
# submodule files to the outside world.
__all__ = stopwatch.__all__, timer.__all__


# Import submodule files so
# classes and functions are usable at
# 'from jutl.pipelining import _' level.
from .datapipeline import *
from .instructionpipeline import *

# Only show functions specified in
# submodule files to the outside world.
__all__ = datapipeline.__all__, instructionpipeline.__all__



# Import submodule files so
# classes and functions are usable at
# 'from jutl.datastructures import _' level.
from .queue import *
from .stack import *

# Only show functions specified in
# submodule files to the outside world.
__all__ = stack.__all__


# Module imports
from jutl.exceptions import EmptyPipelineError, MissingInputError

# External class visibility
__all__ = ['InstructionPipeline']


class InstructionPipeline():
    """
    Class which implements an instruction
    pipeline which by connecting the data
    flow between passed functions.
    """
    def __init__(self, *functions: object, name: str = None):
        "Initialization method."
        self.name: str = name
        self.functions: list[object] = list(functions)

    def __repr__(self) -> str:
        """
        Tells the interpreter how
        to represent this class.
        """
        if self.name is None:
            if len(self.functions) < 1:
                return f"InstructionPipeline()"
            else:
                return f"InstructionPipeline({len(self.functions)})"
        else:
            if len(self.functions) >= 1:
                return f"InstructionPipeline({self.name}, {len(self.functions)})"
            else:
                return f"InstructionPipeline({self.name})"
    
    def __call__(self, data = None) -> None:
        """
        Executes the pipeline
        with passed data.
        """
        if data is None:
            raise MissingInputError("No input data passed to instruction pipeline.")
        if len(self.functions) < 1:
            raise EmptyPipelineError("No functions added to instruction pipeline.")
        else:
            match type(data).__name__:
                case "int" | "float":
                    for function in self.functions:
                        data = function(data)
                case other:
                    raise TypeError(f"Unknown data type passed to instruction pipeline: {other}.")
            return data

    
    def __len__(self) -> int:
        """
        Tells the interpreter what to
        consider this class' length.
        """
        return len(self.functions)
    
    def __iter__(self) -> object:
        """
        Tells the interpreter what to
        iterate over when iterator methods
        are called on this class.
        """
        return iter(self.functions)


    def add(self, *functions: object):
        """
        Adds a function to the pipeline.
        """
        for function in functions:
            self.functions.append(function)

    
    def list(self) -> None:
        """
        Lists all functions in the pipeline.
        """
        for function in self.functions:
            print(function.__name__)

    
    def remove(self, *functions: object):
        """
        Removes a function from the pipeline.
        """
        for function in functions:
            if function in self.functions:
                self.functions.remove(function)
            else:
                continue

    
    def clear(self):
        """
        Removes all functions from the pipeline.
        """
        self.functions.clear()



# Module imports
from jutl.pipelining import InstructionPipeline
from jutl.timers import Timer

# External class visibility
__all__ = ['Latency']


class Latency():
    """
    Class which calculates the latency
    of a single function or method call.
    """
    def __init__(self, func: function):
        "Initialization method."
        self.func: function = func
        self.pipeline: InstructionPipeline = InstructionPipeline(name="Latency")
        self.timer: Timer = Timer(name="Latency")
    
    def __call__(self, data: object = None) -> None:
        self.timer.start()
        result = self.func(data)
        self.timer.stop()
        return result



# Module imports
from __future__ import annotations
from typing import Any
from jutl.exceptions import EmptyStackError, FullStackError
from jutl.formatting import apply

# External class visibility
__all__ = ['Stack']


class Stack(object):
    """
    Class which implements a
    first-in-last-out stack object
    with stack methods.
    """
    def __init__(self, name: str = None, capacity: int = None) -> None:
        "Initialization method."
        self.name: str = name
        self._stack: list = []
        self._capacity: int = capacity

    def __repr__(self) -> str:
        """
        Tells the interpreter how
        to represent this class.
        """
        if self.name is None:
            if len(self) == 0:
                return "Stack()"
            else:
                return f"Stack({self._stack})"

        else:
            if len(self) == 0:
                return f"Stack({self.name})"
            else:
                return f"Stack({self.name}, {self._stack})"
        
    def __call__(self) -> list:
        """
        Tells the interpreter what to
        do when an object of this
        class is called directly.
        """
        return self._stack
    
    def __len__(self) -> int:
        """
        Tells the interpreter what to
        consider this class' length.
        """
        return len(self._stack)
    
    def __iter__(self) -> iter:
        """
        Tells the interpreter what to
        iterate over when iterator methods
        are called on this class.
        """
        raise NotImplementedError("Stacks are not iterable.")
    
    def __eq__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles equal operators.
        """
        return self._stack == other._stack
    
    def __ne__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles not equal operators.
        """
        return self._stack != other._stack

    def __gt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater than operators.
        """
        return len(self) > len(other)

    def __ge__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater or equal operators.
        """
        return len(self) >= len(other)

    def __lt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than operators.
        """
        return len(self) < len(other)

    def __le__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than or equal operators.
        """
        return len(self) <= len(other)

    def __add__(self, other) -> list:
        """
        Tells the interpreter how to sum these objects.
        """
        return self.extend(other=other)


    def push(self, *args: object) -> None:
        """
        Pushes an item onto the stack.
        """
        for item in args:
            if self.is_full:
                raise FullStackError("The stack is full.")
            else:
                self._stack.append(item)


    def pop(self) -> Any:
        """
        Removes the item at the top of the stack.
        """
        if self.is_empty:
            raise EmptyStackError("The stack is empty.")
        else:
            popped = self.top
            self._stack.pop(-1)
            return popped
    

    def clear(self) -> None:
        """
        Clears the stack.
        """
        self._stack.clear()


    @property
    def top(self) -> Any:
        """
        Returns the item at the top of
        the stack without popping it.
        """
        if self.is_empty:
            raise EmptyStackError("The stack is empty.")
        else:
            return self._stack[-1]
        
    
    @property
    def bottom(self) -> Any:
        """
        Returns the item at the bottom
        of the stack without popping it.
        """
        if self.is_empty:
            raise EmptyStackError("The stack is empty.")
        else:
            return self._stack[0]


    @property
    def is_empty(self) -> bool:
        """
        Returns whether the stack is empty.
        """
        if len(self) == 0:
            return True
        else:
            return False


    @property
    def is_full(self) -> bool:
        """
        Returns whether the stack is full.
        """
        if self._capacity is not None:
            if len(self) >= self._capacity:
                return True
            else:
                return False
        else:
            return False


    def extend(self, other: Stack) -> Stack:
        """
        Extends this stack with another stack.
        """
        for item in other._stack:
            self.push(item)
        return self
    


# Module imports
from jutl.formatting import apply
from time import time

# External class visibility
__all__ = ['Stopwatch']


class Stopwatch():
    """
    Class which acts as a stopwatch
    with time and lap methods.
    """
    def __init__(self, name: str = None):
        "Initialization method."
        self.name: str = name
        self._start_time: float
        self._stop_time: float
        self.total_time: float = None
        self._laps: list[float] = []
        self.lap_times: list[float] = []

    def __repr__(self) -> str:
        """
        Tells the interpreter how
        to represent this class.
        """
        if self.name is None:
            if self.total_time is None:
                return "Stopwatch()"
            else:
                return f"Stopwatch({round(self.total_time, 2)}s)"
        else:
            if self.total_time is None:
                return f"Stopwatch({self.name})"
            else:
                return f"Stopwatch({self.name}, {round(self.total_time, 2)}s)"

    def __call__(self, color: str = None):
        """
        Tells the interpreter what to
        do when an object of this
        class is called directly.
        """
        if self.lap_times:
            for n, time in enumerate(self.lap_times):
                if color:
                    print(apply(text=f"Lap {n+1}: {round(time, 2)}s", text_color=color))
                else:
                    print(f"Lap {n+1}: {round(time, 2)}s")
        else:
            print("There are no lap times.")

    def __len__(self) -> int:
        """
        Tells the interpreter what to
        consider this class' length.
        """
        return len(self._laps)

    def __iter__(self) -> iter:
        """
        Tells the interpreter what to
        iterate over when iterator methods
        are called on this class.
        """
        return iter(self.lap_times)

    def __eq__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles equal operators.
        """
        return self.total_time == other.total_time

    def __ne__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles not equal operators.
        """
        return self.total_time != other.total_time

    def __gt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater than operators.
        """
        return self.total_time > other.total_time

    def __ge__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater or equal operators.
        """
        return self.total_time >= other.total_time

    def __lt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than operators.
        """
        return self.total_time < other.total_time

    def __le__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than or equal operators.
        """
        return self.total_time <= other.total_time

    def __add__(self, other) -> float:
        """
        Tells the interpreter how to sum these objects.
        """
        return self.total_time + other.total_time

    def __sub__(self, other) -> float:
        """
        Tells the interpreter how to subtract these objects.
        """
        return self.total_time - other.total_time

    def __mul__(self, multiplier) -> float:
        """
        Tells the interpreter how to subtract these objects.
        """
        return self.total_time * multiplier

    def __truediv__(self, other) -> float:
        """
        Tells the interpreter how to subtract these objects.
        """
        return self.total_time / other.total_time


    def start(self, message: str = None, color: str = None):
        """
        Starts the stopwatch and
        optionally prints a message.
        """
        self._start_time = time()
        if message:
            print(apply(text=message, text_color=color) if color else print(message))


    def lap(self, lap_time: float = None, message: str = None, color: str = None):
        """
        Adds the current time to the lap
        time list, records the time since
        the start or time of the last recorded
        lap, and optionally prints a message.
        """
        if lap_time:
            self._laps.append(lap_time)
        else:
            self._laps.append(time())
        if not self.lap_times:
            self.lap_times.append(self._calculate_time(self._start_time, self._laps[-1]))
        else:
            self.lap_times.append(self._calculate_time(self._laps[-2], self._laps[-1]))
        if message:
            print(apply(text=message, text_color=color) if color else print(message))


    def stop(self, message: str = None, color: str = None):
        """
        Stops the stopwatch, calculates
        the total time passed, and
        optionally prints a message.
        """
        self._stop_time = time()
        self.total_time = self._calculate_time(self._start_time, self._stop_time)
        self.lap(self._stop_time) # add final lap time and calculate lap difference
        if message:
            print(apply(text=message, text_color=color) if color else print(text=message))


    def _calculate_time(self, time1: float, time2: float) -> float:
        """
        Returns the difference in time where t2>t1.
        """
        return time2 - time1


    def reset(self, message: str = None, color: str = None):
        """
        Resets all stopwatch attributes
        and optionally prints a message.
        """
        self._start_time = None
        self._stop_time = None
        self.total_time = None
        self._laps.clear()
        self.lap_times.clear()
        if message:
            print(apply(text=message, text_color=color) if color else print(text=message))



# Import submodule files so
# classes and functions are usable at
# 'from jutl.exceptions import _' level.
from .emptypipeline import *
from .emptyqueue import *
from .emptystack import *
from .fullqueue import *
from .fullstack import *
from .invalidformatting import *
from .missinginput import *

# Only show functions specified in
# submodule files to the outside world.
__all__ = emptypipeline.__all__, emptyqueue.__all__, emptystack.__all__, fullqueue.__all__, fullstack.__all__, invalidformatting.__all__, missinginput.__all__



# Import submodule files so
# classes and functions are usable at
# 'from jutl.benchmarking import _' level.
from .latency import *

# Only show functions specified in
# submodule files to the outside world.
__all__ = latency.__all__


# Benchmarks:
# - Timed function call: Latency

# - Generic benchmamrk class: Benckmark
# - Timed pipeline with single data input (InstructionPipeline): Benchmark
# - Timed pipeline with multiple data inputs & flippable loops (DataPipeline): 





# Module imports
from __future__ import annotations
from typing import Any
from jutl.exceptions import EmptyQueueError, FullQueueError
from jutl.formatting import apply

# External class visibility
__all__ = ['Queue']


class Queue(object):
    """
    Class which implements a
    first-in-first-out queue object
    with queue methods.
    """
    def __init__(self, name: str = None, capacity: int = None) -> None:
        "Initialization method."
        self.name: str = name
        self._queue : list = []
        self._capacity: int = capacity

    def __repr__(self) -> str:
        """
        Tells the interpreter how
        to represent this class.
        """
        string = f"{self.__class__.__name__}("
        if self.name is not None:
            string += f"{self.name}"
        if len(self._queue) > 0:
            string += f", " if self.name is not None else ""
            string += f"{len(self._queue)}"
        if self._capacity is not None:
            string += f"/" if len(self) > 0 else ", " if self.name is not None else ""
            string += f"{self._capacity}"
        string += ")"
        return string

    def __call__(self) -> list:
        """
        Tells the interpreter what to
        do when an object of this
        class is called directly.
        """
        return self._queue

    def __len__(self) -> int:
        """
        Tells the interpreter what to
        consider this class' length.
        """
        return len(self._queue)

    def __iter__(self) -> iter:
        """
        Tells the interpreter what to
        iterate over when iterator methods
        are called on this class.
        """
        raise NotImplementedError("Queues are not iterable.")

    def __eq__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles equal operators.
        """
        return self._queue == other._queue

    def __ne__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles not equal operators.
        """
        return self._queue != other._queue

    def __gt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater than operators.
        """
        return len(self) > len(other)

    def __ge__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles greater or equal operators.
        """
        return len(self) >= len(other)

    def __lt__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than operators.
        """
        return len(self) < len(other)

    def __le__(self, other) -> bool:
        """
        Tells the interpreter how this class
        handles less than or equal operators.
        """
        return len(self) <= len(other)

    def __add__(self, other) -> list:
        """
        Tells the interpreter how to sum these objects.
        """
        return self.extend(other=other)


    def enqueue(self, *args: object) -> None:
        """
        Adds an item to the back of the queue.
        """
        for item in args:
            if self.is_full:
                raise FullQueueError("The queue is full.")
            else:
                self._queue.append(item)


    def dequeue(self) -> Any:
        """
        Removes the item from at front of the queue.
        """
        if self.is_empty:
            raise EmptyQueueError("The queue is empty.")
        else:
            dequeued = self.front
            self._queue.pop(0)
            return dequeued
        

    def clear(self) -> None:
        """
        Clears the queue.
        """
        self._queue.clear()


    @property
    def front(self) -> Any:
        """
        Returns the item at the front of
        the queue without dequeueing it.
        """
        if self.is_empty:
            raise EmptyQueueError("The queue is empty.")
        else:
            return self._queue[0]


    @property
    def rear(self) -> Any:
        """
        Returns the item at the rear of
        the queue without dequeueing it.
        """
        if self.is_empty:
            raise EmptyQueueError("The queue is empty.")
        else:
            return self._queue[-1]


    @property
    def is_empty(self) -> bool:
        """
        Returns whether the queue is empty.
        """
        if len(self) <= 0:
            return True
        else:
            return False


    @property
    def is_full(self) -> bool:
        """
        Returns whether the queue is full.
        """
        if self._capacity is not None:
            if len(self) >= self._capacity:
                return True
            else:
                return False
        else:
            return False


    def extend(self, other: Queue) -> Queue:
        """
        Extends this queue with another queue.
        """
        for item in other._queue:
            self.enqueue(item)
        return self
    


# External function visibility
__all__ = ['hello_world']

def hello_world(name: str = None) -> str:
    """
    Greets the user.

    Returns "Hello, 'user!'" where 'user' is 'name' if passed,
    or "Hello, World!" if 'name' is not passed.

            Parameters:
                    name (str): The user's name
            
            Returns:
                    str: A greeting for the user
    """
    if name:
        return(f"Hello, {name}!")
    else:
        return(f"Hello, World!")



# Module imports
from setuptools import setup

# Arguments
git_name = "jutils"
pypi_name = "jutl"
version = "0.5.2" # update __init__.py
python_version = ">=3.10"

# Long description from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

# jutils package data
jutils_package_data = [
    'averages/*',
    'calculators/*',
    'converters/*',
    'cryptography/*',
    'datastructures/*',
    'formatting/*',
    'language/*',
    'logic/*',
    'pipelining/*',
    'sorters/*',
    'timers/*',
    'utilities/*'
]

# Run setup function
setup(
    name=pypi_name,
    version=version,
    description='A Python package of useful tools and utilities.',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jordan Welsman',
    author_email='jordan.welsman@outlook.com',
    url='https://pypi.org/project/'+pypi_name+'/',
    download_url='https://github.com/JordanWelsman/jutils/tags',
    package_data={f'{pypi_name}': jutils_package_data},
    python_requires=python_version,
    # jutils package information
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11"
    ],
    keywords='python, averages, calculators, converters, cryptography, formatting, language, logic, pipelining, sorters, timers, utilities'
)


# Module imports
from jutl.pipelining.instructionpipeline import InstructionPipeline
from jutl.exceptions import EmptyPipelineError, MissingInputError

# External class visibility
__all__ = ['DataPipeline']


class DataPipeline(InstructionPipeline):
    """
    Class which extends the instruction
    pipeline and supports passing
    multiple instances of data.
    """
    def __repr__(self) -> str:
        """
        Tells the interpreter how
        to represent this class.
        """
        if self.name is None:
            if len(self.functions) < 1:
                return f"DataPipeline()"
            else:
                return f"DataPipeline({len(self.functions)})"
        else:
            if len(self.functions) >= 1:
                return f"DataPipeline({self.name}, {len(self.functions)})"
            else:
                return f"DataPipeline({self.name})"

    def __call__(self, *data):
        """
        Executes the pipeline
        with passed data.
        """
        results: list = []
        if len(data) < 1:
            raise MissingInputError("No input data passed to data pipeline.")
        if len(self.functions) < 1:
            raise EmptyPipelineError("No functions added to data pipeline.")
        for datum in data:
            match type(datum).__name__:
                case "int" | "float":
                    for function in self.functions:
                        datum = function(datum)
                    results.append(datum)
                case other:
                    raise TypeError(f"Unknown data type passed to instruction pipeline: {other}.")
        return results
